"""Module to test API class."""

import pytest

from mailerlite import MailerLiteApi
from mailerlite.constants import API_KEY_TEST


@pytest.fixture
def header():
    headers = {'content-type': "application/json",
               "X-MailerLite-ApiDocs": "true",
               'x-mailerlite-apikey': API_KEY_TEST
               }
    return headers


def test_wrong_headers(header):
    # test valid first
    try:
        _ = MailerLiteApi(header)
    except ValueError:
        return

    headers_2 = {'content-type': "application/json",
                 "X-MailerLite-ApiDocs": "true",
                 'x-mailerlite-apikey': 'FAKE_KEY'
                 }

    with pytest.raises(ValueError):
        api = MailerLiteApi(headers_2)

    with pytest.raises(OSError):
        api = MailerLiteApi('FAKE_KEY')
        api.subscribers.all()


def test_api(header):
    try:
        api = MailerLiteApi(API_KEY_TEST)
    except ValueError:
        return

    assert api.headers == header

    batch_requests = {"requests": [{"method": "GET",
                                    "path": "/api/v2/groups"
                                    },
                                   {"method": "GET",
                                    "path": "/api/v2/fields"
                                    }
                                   ]
                      }
    res = api.batch(batch_requests)
    assert len(res) == 2



"""Manage Subcribers."""

from warnings import warn
import mailerlite.client as client
from mailerlite.constants import Subscriber, Activity, Group, Field


def get_id_or_email_identifier(**kwargs):
    """Return id or email identifier."""
    identifier = None
    for k in ['id', 'email']:
        identifier = kwargs.get(k, None)
        if identifier is not None:
            break
    return identifier


class Subscribers:

    def __init__(self, headers):
        """Initialize Subscribers object.

        Parameters
        ----------
        headers : dict
            request header containing your mailerlite api_key.
            More information : https://developers.mailerlite.com/docs/request

        """
        valid_headers, error_msg = client.check_headers(headers)
        if not valid_headers:
            raise ValueError(error_msg)

        self.headers = headers

    def active(self, limit=100, offset=0, as_json=False):
        """Get all active Subscribers from your account.

        look at https://developers.mailerlite.com/v2/reference#subscribers

        Parameters
        ----------
        limit : int
            How many subscribers you want
        offset : int
            page index
        as_json : bool
            return result as json format

        Returns
        -------
        subscribers: list
            all active Subscribers. More informations :
            https://developers.mailerlite.com/v2/reference#subscribers
        """
        return self.all(limit=limit, offset=offset, stype='active',
                        as_json=as_json)

    def unsubscribed(self, limit=100, offset=0, as_json=False):
        """Get all unsubscribed Subscribers from your account.

        look at https://developers.mailerlite.com/v2/reference#subscribers

        Parameters
        ----------
        limit : int
            How many subscribers you want
        offset : int
            page index
        as_json : bool
            return result as json format

        Returns
        -------
        subscribers: list
            all unsubscribed Subscribers. More informations :
            https://developers.mailerlite.com/v2/reference#subscribers
        """
        return self.all(limit=limit, offset=offset, stype='unsubscribed',
                        as_json=as_json)

    def bounced(self, limit=100, offset=0, as_json=False):
        """Get all bounced Subscribers from your account.

        look at https://developers.mailerlite.com/v2/reference#subscribers

        Parameters
        ----------
        limit : int
            How many subscribers you want
        offset : int
            page index
        as_json : bool
            return result as json format

        Returns
        -------
        subscribers: list
            all bounced Subscribers. More informations :
            https://developers.mailerlite.com/v2/reference#subscribers
        """
        return self.all(limit=limit, offset=offset, stype='bounced',
                        as_json=as_json)

    def junk(self, limit=100, offset=0, as_json=False):
        """Get all junk Subscribers from your account.

        look at https://developers.mailerlite.com/v2/reference#subscribers

        Parameters
        ----------
        limit : int
            How many subscribers you want
        offset : int
            page index
        as_json : bool
            return result as json format

        Returns
        -------
        subscribers: list
            all junk Subscribers. More informations :
            https://developers.mailerlite.com/v2/reference#subscribers
        """
        return self.all(limit=limit, offset=offset, stype='junk',
                        as_json=as_json)

    def unconfirmed(self, limit=100, offset=0, as_json=False):
        """Get all unconfirmed Subscribers from your account.

        look at https://developers.mailerlite.com/v2/reference#subscribers

        Parameters
        ----------
        limit : int
            How many subscribers you want
        offset : int
            page index
        as_json : bool
            return result as json format

        Returns
        -------
        subscribers: list
            all unconfirmed Subscribers. More informations :
            https://developers.mailerlite.com/v2/reference#subscribers
        """
        return self.all(limit=limit, offset=offset, stype='unconfirmed',
                        as_json=as_json)

    def all(self, limit=100, offset=0, stype=None, as_json=False):
        """Get paginated details of all Subscribers from your account.

        look at https://developers.mailerlite.com/v2/reference#subscribers

        Parameters
        ----------
        limit : int
            How many subscribers you want
        offset : int
            page index
        stype : str
            Define subscriber type: Here are the possible values:
            * None - All subscribers (default)
            * active
            * unsubscribed
            * bounced
            * junk
            * unconfirmed
        as_json : bool
            return result as json format

        Returns
        -------
        subscribers: list
            all desired Subscribers. More informations :
            https://developers.mailerlite.com/v2/reference#subscribers
        """
        params = {'limit': limit, 'offset': offset}
        if stype and stype.lower() in ['active', 'unsubscribed', 'bounced',
                                       'junk', 'unconfirmed']:
            params.update({'type': stype})

        url = client.build_url('subscribers', **params)
        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        for res in res_json:
            res['fields'] = [Field(**field) for field in res['fields']]

        all_subscribers = [Subscriber(**res) for res in res_json]
        return all_subscribers

    def count(self, stype=None, as_json=False):
        """Get the count of subscribers of a type.

        Please, be aware that `count` is not documented in the official API
        https://developers.mailerlite.com/reference#subscribers

        Parameters
        ----------
        stype : str
            Define subscriber type: Here are the possible values:
            * None - All subscribers (default)
            * active
            * unsubscribed
            * bounced
            * junk
            * unconfirmed
        as_json : bool
            return result as json format

        Returns
        -------
        number: Integer
            the count of subscribers
        """
        warn("Please, be aware that `count` is not in the official API")

        params = {}
        if stype and stype.lower() in ['active', 'unsubscribed', 'bounced',
                                       'junk', 'unconfirmed']:
            params.update({'type': stype})

        url = client.build_url('subscribers', 'count', **params)
        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        return res_json['count']

    def get(self, as_json=False, **identifier):
        """Get a single subscriber from your account.

        https://developers.mailerlite.com/v2/reference#single-subscriber

        Parameters
        ----------
        identifier : str
            should be subscriber id or email.
            e.g: id=1343965485 or email='demo@mailerlite.com'
        as_json : bool
            return result as json format
        Returns
        -------
        subscriber: :class:Subscriber
            a single subscriber
        """
        path = get_id_or_email_identifier(**identifier)
        if path is None:
            raise IOError('An identifier must be define')

        url = client.build_url('subscribers', path)
        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        res_json['fields'] = [Field(**res) for res in res_json['fields']]

        return Subscriber(**res_json)

    def delete(self, subscriber_id):
        """Remove a subscribers.

        Parameters
        ----------
        subscriber_id : int
            subscribers id

        Returns
        -------
        success: bool
            deletion status
        """
        url = client.build_url('subscribers', subscriber_id)
        return client.delete(url, headers=self.headers)

    def search(self, search=None, limit=100, offset=0, minimized=True,
               as_json=False):
        """Get paginated details of all Subscribers from your account.

        look at https://developers.mailerlite.com/v2/reference#subscribers

        Parameters
        ----------
        search : str
            query parameter to search
        limit : int
            How many subscribers you want
        offset : int
            page index
        minimized : bool
            return minimized response with: id, email, type
            default: True
        as_json : bool
            return result as json format

        Returns
        -------
        subscribers: list
            all desired Subscribers. More informations :
            https://developers.mailerlite.com/v2/reference#subscribers
        """
        params = {'limit': limit, 'offset': offset, 'minimized': minimized}
        if search is not None:
            params.update({'query': search})
        url = client.build_url('subscribers', 'search', **params)

        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        if not minimized:
            for res in res_json:
                res['fields'] = [Field(**field) for field in res['fields']]

        all_subscribers = [Subscriber(**res) for res in res_json]
        return all_subscribers

    def groups(self, as_json=False, **identifier):
        """Get groups subscriber belongs to.

        More informations:
        https://developers.mailerlite.com/v2/reference#groups-subscriber-belongs-to

        Parameters
        ----------
        identifier : str
            should be subscriber id or email.
            e.g: id=1343965485 or email='demo@mailerlite.com'
        as_json : bool
            return result as json format

        Returns
        -------
        groups: list
            all groups that a subscriber belongs to. More informations :
            https://developers.mailerlite.com/v2/reference#groups
        """
        path = get_id_or_email_identifier(**identifier)
        if path is None:
            raise IOError('An identifier must be define')

        url = client.build_url('subscribers', path, 'groups')

        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        all_groups = [Group(**res) for res in res_json]
        return all_groups

    def activity(self, as_json=False, atype=None, limit=100,
                 offset=0, **identifier):
        """Get activities (clicks, opens, etc) of selected subscriber.

        More informations:
        https://developers.mailerlite.com/v2/reference#activity-of-single-subscriber

        Parameters
        ----------
        identifier : str
            should be subscriber id or email.
            e.g: id=1343965485 or email='demo@mailerlite.com'
        as_json : bool
            return result as json format
        atype : str
            Define activity type: Here are the possible values:
            * None - All activities (default)
            * opens
            * clicks
            * bounces
            * junks
            * unsubscribes
            * forwards
            * sendings
        limit : int, optional
            How many subscribers you want, default 100
        offset : int, optional
            page index, default 0

        Returns
        -------
        activities: list
            all subscriber activities. More informations :
            https://developers.mailerlite.com/v2/reference#activity-of-single-subscriber
        """
        params = {'limit': limit, 'offset': offset}
        path = get_id_or_email_identifier(**identifier)
        if path is None:
            raise IOError('An identifier must be define')

        args = ['subscribers', path, 'activity']
        if atype:
            possible_atype = ['opens', 'clicks', 'junks', 'bounces',
                              'unsubscribes', 'forwards', 'sendings']
            if atype not in possible_atype:
                raise ValueError('Incorrect value atype. Activity type should'
                                 ' be {0}'.format(possible_atype))
            args.append(atype)

        url = client.build_url(*args, **params)

        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        all_activities = [Activity(**res) for res in res_json]
        return all_activities

    def update(self, data, as_json=False, **identifier):
        """Update single subscriber.

        https://developers.mailerlite.com/v2/reference#update-subscriber

        Parameters
        ----------
        data : dict
            subscriber object. only the email is required.
            you can use the following example:
            data = {'name'   : 'John',
                    'fields' : {'company': 'MailerLite'}
                    }
        as_json : bool
            return result as json format
        identifier : str
            should be subscriber id or email.
            e.g: id=1343965485 or email='demo@mailerlite.com'

        Returns
        -------
        response : int
            response value
        content : dict
            The JSON output from the API

        Notes
        -----
        The email of a subscriber can not be updated

        """
        path = get_id_or_email_identifier(**identifier)
        if path is None:
            raise IOError('An identifier must be define')

        if 'email' in data.keys():
            raise ValueError("Subscriber email can not be updated. Please, "
                             "remove this field or create a new Subscriber. "
                             "For more informations, look at "
                             "http://help.mailerlite.com/article/show"
                             "/29233-how-to-edit-a-subscribers-data")

        optional_keys = ['name', 'type', 'fields', 'resend_autoresponders']
        unknown_keys = [d for d in data.keys() if d not in optional_keys
                        if d not in ['groups', 'segments']]
        if unknown_keys:
            raise ValueError("The following keys are unknown: {}"
                             .format(unknown_keys))

        url = client.build_url('subscribers', path)
        _, res_json = client.put(url, body=data, headers=self.headers)

        if as_json or not res_json:
            return res_json

        res_json['fields'] = [Field(**res) for res in res_json['fields']]

        return Subscriber(**res_json)

    def create(self, data, as_json=False):
        """Add new single subscriber.

        https://developers.mailerlite.com/v2/reference#create-a-subscriber

        Parameters
        ----------
        data : dict
            subscriber object. only the email is required.
            you can use the following example:
            data = {'name'   : 'John',
                    'email'  : 'demo@mailerlite.com',
                    'fields' : {'company': 'MailerLite'}
                    }
        as_json : bool
            return result as json format
        Returns
        -------
        subscriber: :class:Subscriber
            a single subscriber
        """
        if not isinstance(data, dict):
            raise ValueError('In data should be a dictionary.')
        required_keys = ['email', ]
        optional_keys = ['name', 'fields', 'resubscribe', 'type',
                         'signup_ip', 'signup_timestamp', 'confirmation_ip',
                         'confirmation_timestamp']
        available_keys = required_keys + optional_keys

        errors = [rk for rk in required_keys if rk not in data.keys()]
        if errors:
            raise ValueError("The following keys are missing and they"
                             " are required : {}".format(errors))

        unknown_keys = [d for d in data.keys() if d not in available_keys
                        if d not in ['groups', 'segments']]
        if unknown_keys:
            raise ValueError("The following keys are unknown: {}"
                             .format(unknown_keys))

        url = client.build_url('subscribers')
        _, res_json = client.post(url, body=data, headers=self.headers)

        if as_json or not res_json:
            return res_json

        return Subscriber(**res_json)



"""File containing global contants ."""

from collections import namedtuple

MAILERLITE_API_V2_URL = 'https://api.mailerlite.com/api/v2/'

API_KEY_TEST = "fc7b8c5b32067bcd47cafb5f475d2fe9"

VALID_REQUEST_METHODS = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']

Field = namedtuple('Field', ['key', 'value', 'type', 'title', 'id',
                             'date_updated', 'date_created'])
Group = namedtuple('Group', ["id", "name", "total", "active", "unsubscribed",
                             "bounced", "unconfirmed", "junk", "sent",
                             "opened", "clicked", "date_created",
                             "date_updated", "parent_id"])
Activity = namedtuple('Activity', ['date', 'report_id', 'subject',
                                   'campaign_name', 'type',
                                   'campaign_id', 'link_id', 'link',
                                   'receiver', 'receiver_name',
                                   'receiver_email', 'sender', 'sender_name',
                                   'sender_email'])
Subscriber = namedtuple('Subscriber', ['id', 'name', 'email', 'sent',
                                       'opened', 'clicked', 'type',
                                       'signup_ip', 'signup_timestamp',
                                       'confirmation_ip',
                                       'confirmation_timestamp',
                                       'fields', 'date_subscribe',
                                       'subscribe_date',
                                       'webform_subscribe_date',
                                       'date_unsubscribe', 'date_created',
                                       'date_updated', 'opened_rate',
                                       'clicked_rate', 'country_id',
                                       'user_agent'
                                       ])
Pagination = namedtuple('Pagination', ["total", "count", "per_page",
                                       "current_page", "total_pages",
                                       "links"])
Meta = namedtuple('Meta', ['pagination', ])
Segment = namedtuple('Segment', ['id', 'title', 'filter', 'total', 'sent',
                                 'opened', 'clicked', 'created_at',
                                 'updated_at', 'timed_out'])
Stats = namedtuple('Stats', ['count', 'rate'])
Campaign = namedtuple('Campaign', ['id', 'total_recipients', 'type',
                                   'date_created', 'date_send', 'name',
                                   'subject', 'status',
                                   'opened', 'clicked'])
Webhook = namedtuple('Webhook', ['id', 'event', 'url', 'created_at',
                                 'updated_at'])

for nt in [Subscriber, Field, Group, Activity, Segment, Meta, Pagination,
           Campaign, Stats, Webhook]:
    nt.__new__.__defaults__ = (None,) * len(nt._fields)



"""Manage Campaign."""

import mailerlite.client as client
from mailerlite.constants import Campaign, Stats


class Campaigns:

    def __init__(self, headers):
        """Initialize Campaigns object.

        Parameters
        ----------
        headers : dict
            request header containing your mailerlite api_key.
            More information : https://developers.mailerlite.com/docs/request
        """
        valid_headers, error_msg = client.check_headers(headers)
        if not valid_headers:
            raise ValueError(error_msg)

        self.headers = headers

    def all(self, status='sent', limit=100, offset=0, order='asc',
            as_json=False):
        """Get paginated details of all campaigns from your account.

        look at https://developers.mailerlite.com/reference#campaigns-by-type

        Parameters
        ----------
        status : str
            Define campaigns type: Here are the possible values:
            * sent - campaigns which are sent already (default)
            * draft - campaigns which aren't completed or sent to subscribers
            * outbox - campaigns which are being sent right now or scheduled
        limit : int
            How many campaigns you want
        offset : int
            page index
        order : str
            pick the order. Here are the possible values: ASC (default) or DESC
        as_json : bool
            return result as json format

        Returns
        -------
        campaigns: list
            all desired campaign. More informations concerning campaign :
            https://developers.mailerlite.com/reference#section-response-body-parameters
        """
        if order.upper() not in ['ASC', 'DESC']:
            raise IOError("Incorrect order, please choose between ASC or DESC")

        params = {'limit': limit, 'offset': offset, 'order': order}
        url = client.build_url('campaigns', status, **params)
        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        for res in res_json:
            res['opened'] = Stats(**res.pop('opened'))
            res['clicked'] = Stats(**res.pop('clicked'))

        all_campaigns = [Campaign(**res) for res in res_json]
        return all_campaigns

    # def get(self, campaign_id, as_json=False):
    #     """ NOT AVAILABLE via API
    # Get a campaigns from your account.

    #     look at https://developers.mailerlite.com/reference#campaigns-by-type

    #     Parameters
    #     ----------
    #     campaign_id : int
    #         should be campaign id. e.g: id=1343965485
    #     as_json : bool
    #         return result as json format
    #     Returns
    #     -------
    #     campaign: :class:`Campaign`
    #         a single campaign
    #     """
    #     url = client.build_url('campaigns', campaign_id)
    #     res_code, res_json = client.get(url, headers=self.headers)

    #     if as_json or not res_json:
    #         return res_json

    #     res_json['opened'] = Stats(**res_json.pop('opened'))
    #     res_json['clicked'] = Stats(**res_json.pop('clicked'))

    #     return Campaign(**res_json)

    def update(self, campaign_id, html, plain, auto_inline=True):
        r"""Upload your HTML template to created campaign.

        https://developers.mailerlite.com/v2/reference#put-custom-content-to-campaign

        Parameters
        ----------
        campaign_id : int
            the campaign id that you want to update
        html : str
            HTML template source
        plain : str
            Plain text of email
        auto_inline : bool, optional
            Defines if it is needed to convert available CSS to inline CSS
            (excluding media queries)

        Returns
        -------
        success : bool

        Examples
        --------
        >>> from mailerlite import MailerLiteApi
        >>> api = MailerLiteApi('my_keys')
        >>> html = '<head></head><body><h1>Title</h1><p>Content</p><p><small>'
        >>> html += '<a href=\"{$unsubscribe}\">Unsubscribe</a></small></p>'
        >>> html += '</body>'
        >>> plain = "Your email client does not support HTML emails. "
        >>> plain += "Open newsletter here: {$url}. If you do not want"
        >>> plain += " to receive emails from us, click here: {$unsubscribe}"
        >>> api.campaigns.update(campaign_id, html=html, plain=plain)
        True

        Notes
        -----
        * HTML template must contain body and head tag.
        * HTML template must contain a link for unsubscribe. It may look like
            this: <a href="{$unsubscribe}">Unsubscribe</a>
        * Some email clients do not support HTML emails so you need to set
            plain text email and it must contain these variables:
            * {$unsubscribe} - unsubscribe link
            * {$url} - URL to your HTML newsletter

        """
        url = client.build_url('campaigns', campaign_id, 'content')
        # Todo, Check html syntax
        body = {"html": html, "plain": plain}
        _, res_json = client.put(url, body=body, headers=self.headers)

        if not res_json:
            return False

        return res_json['success']

    def create(self, data):
        """Create campaign where you will use your custom HTML template.

        https://developers.mailerlite.com/v2/reference#campaigns

        Parameters
        ----------
        data : dict
            campaign object. For a regular campaign, you  can use the
            following example:
            data = {"subject": "Regular campaign subject",
                    "name": "Regular campaign name",
                    "groups": [2984475, 3237221],
                    "type": "regular"}
            For a/b  campaign type, you need more fields, Here an example:
            data = {"groups": [2984475, 3237221],
                    "type": "ab",
                    "ab_settings": {"send_type": "subject",
                                    "values": ["Email subject A",
                                               "Email subject B"],
                                    "ab_win_type": "opens",
                                    "winner_after": 1,
                                    "winner_after_type": "h",
                                    "split_part": "10"
                                }}

        Returns
        -------
        response : int
            response value
        content : dict
            The JSON output from the API
        """
        if not isinstance(data, dict):
            raise ValueError('In data should be a dictionary.')
        required_keys = ['type', ]
        optional_keys = ['subject', 'name', 'from', 'from_name', 'language',
                         'ab_settings']
        ab_settings_keys = ['values', 'send_type', 'ab_win_type',
                            'winner_after', 'winner_after_type',
                            'split_part']
        available_keys = required_keys + optional_keys

        errors = [rk for rk in required_keys if rk not in data.keys()]
        if errors:
            raise ValueError("The following keys are missing and they"
                             " are required : {}".format(errors))

        if ('groups' not in data.keys()) and ('segments' not in data.keys()):
            raise ValueError("'groups' key is required if 'segments' key"
                             " are not specified. If specified, 'groups'"
                             " are ignored")

        unknown_keys = [d for d in data.keys() if d not in available_keys
                        if d not in ['groups', 'segments']]
        if unknown_keys:
            raise ValueError("The following keys are unknown: {}"
                             .format(unknown_keys))

        if 'ab_settings' in data.keys():
            errors = [rk for rk in ab_settings_keys
                      if rk not in data['ab_settings'].keys()]
            if errors:
                raise ValueError("The following keys are missing in the"
                                 " ab_settings and they are required : "
                                 "{}".format(errors))

        url = client.build_url('campaigns')
        return client.post(url, body=data, headers=self.headers)

    def send(self, campaign_id):
        """Send out a campaign.

        https://developers.mailerlite.com/reference#campaign-actions-and-triggers

        Parameters
        ----------
        campaign_id : int
            campaign id

        Returns
        -------
        content : dict
            The JSON output from the API

        """
        # TODO: Check if campaign is in Draft otherwise raise an issue
        #  Add parameters like followup / send later / etc...
        url = client.build_url('campaigns', campaign_id, "actions/send")
        return client.post(url, headers=self.headers)

    def cancel(self, campaign_id, as_json=False):
        """Cancel a campaign which is in outbox.

        https://developers.mailerlite.com/reference#campaign-actions-and-triggers

        Parameters
        ----------
        campaign_id : int
            campaign id
        as_json : bool
            return result as json format

        Returns
        -------
        content : dict
            The JSON output from the API

        """
        # TODO: Check if campaign is in Outbox otherwise raise an issue
        url = client.build_url('campaigns', campaign_id, "actions/cancel")
        code, res_json = client.post(url, headers=self.headers)

        # TODO: Check new attribute to campaign object.
        # if as_json or not res_json or code != 200:
        #     return res_json

        # res_json['opened'] = Stats(**res_json.pop('opened'))
        # res_json['clicked'] = Stats(**res_json.pop('clicked'))

        # return Campaign(**res_json)
        return code, res_json

    def delete(self, campaign_id):
        """Remove a campaign.

        look at https://developers.mailerlite.com/reference#delete-campaign

        Parameters
        ----------
        campaign_id : int
            campaign id

        Returns
        -------
        success: bool
            deletion status
        """
        url = client.build_url('campaigns', campaign_id)
        return client.delete(url, headers=self.headers)

    def count(self, status='sent'):
        """Return the number of campaigns.

        Returns
        -------
        status : str
            Define campaigns type: Here are the possible values:
            * sent - campaigns which are sent already (default)
            * draft - campaigns which aren't completed or sent to subscribers
            * outbox - campaigns which are being sent right now or scheduled
        count: int
            number of campaigns
        """
        if status.lower() not in ['sent', 'draft', 'outbox']:
            raise ValueError('Incorrect status, check documentation')
        url = client.build_url('campaigns', status.lower(), 'count')
        _, res_json = client.get(url, headers=self.headers)
        return res_json['count']



"""Utility function for unit tests."""
from functools import wraps
import pytest


def succeed_or_skip_sensitive_tests(func):
    """Small decorator to skip some sensitive function."""
    @wraps(func)
    def wrapper_func(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            print(str(e))
            pytest.skip("Test Failed due to an object deleted "
                        "by another matrix.")

    return wrapper_func



"""Manage Segments."""

import mailerlite.client as client
from mailerlite.constants import Segment, Meta, Pagination


class Segments:

    def __init__(self, headers):
        """Initialize Segments object.

        Parameters
        ----------
        headers : dict
            request header containing your mailerlite api_key.
            More information : https://developers.mailerlite.com/docs/request

        """
        valid_headers, error_msg = client.check_headers(headers)
        if not valid_headers:
            raise ValueError(error_msg)

        self.headers = headers

    def all(self, limit=100, offset=0, order='asc', as_json=False):
        """Get paginated details of all segments from your account.

        look at https://developers.mailerlite.com/reference#segments-1

        Parameters
        ----------
        limit : int
            How many segments you want
        offset : int
            page index
        order : str
            pick the order. Here are the possible values: ASC (default) or DESC
        as_json : bool
            return result as json format

        Returns
        -------
        segments: list
            all desired segment. More informations:
            https://developers.mailerlite.com/v2/reference#segments-1
        meta: object
            some meta informations about the segments

        """
        if order.upper() not in ['ASC', 'DESC']:
            raise IOError("Incorrect order, please choose between ASC or DESC")

        params = {'limit': limit, 'offset': offset, 'order': order}
        url = client.build_url('segments', **params)
        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json['data'], res_json['meta']

        all_segments = [Segment(**res) for res in res_json['data']]
        res_json['meta']['pagination'] = Pagination(**res_json['meta']
                                                    .pop('pagination'))
        meta = Meta(**res_json['meta'])
        return all_segments, meta

    def count(self):
        """Return the number of segments.

        More informations:
        https://developers.mailerlite.com/reference#section-segments-count

        Returns
        -------
        count: int
            number of segments

        """
        url = client.build_url('segments', 'count')
        _, res_json = client.get(url, headers=self.headers)

        return res_json.get('count') or len(res_json.get('data'))



"""Manage Groups."""

import mailerlite.client as client
from mailerlite.constants import Subscriber, Group, Field


class Groups:

    def __init__(self, headers):
        """Initialize Groups object.

        Parameters
        ----------
        headers : dict
            request header containing your mailerlite api_key.
            More information : https://developers.mailerlite.com/docs/request

        """
        valid_headers, error_msg = client.check_headers(headers)
        if not valid_headers:
            raise ValueError(error_msg)

        self.headers = headers

    def all(self, limit=100, offset=0, gfilters='', as_json=False):
        """Get list of groups from your account.

        look at https://developers.mailerlite.com/v2/reference#groups

        Parameters
        ----------
        limit : int
            How many groups you want
        offset : int
            page index
        gfilters : str
            group filters
        as_json : bool
            return result as json format

        Returns
        -------
        groups: list
            all desired Groups. More informations :
            https://developers.mailerlite.com/v2/reference#groups

        """
        params = {'limit': limit, 'offset': offset, 'filters': gfilters}
        url = client.build_url('groups', **params)
        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        all_groups = [Group(**res) for res in res_json]
        return all_groups

    def get(self, group_id, as_json=False):
        """Get single group by ID from your account.

        look at https://developers.mailerlite.com/v2/reference#single-group

        Parameters
        ----------
        group_id : int
            should be group id. e.g: id=1343965485
        as_json : bool
            return result as json format
        Returns
        -------
        Group: :class:Group
            a single group

        """
        url = client.build_url('groups', group_id)
        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        return Group(**res_json)

    def delete(self, group_id):
        """Remove a group.

        https://developers.mailerlite.com/v2/reference#delete-group

        Parameters
        ----------
        group_id : int
            group id

        Returns
        -------
        success: bool
            deletion status

        """
        url = client.build_url('groups', group_id)
        return client.delete(url, headers=self.headers)

    def update(self, group_id, name, as_json=False):
        """Update existing group.

        https://developers.mailerlite.com/v2/reference#rename-group

        Parameters
        ----------
        group_id : int
            group that you want to rename
        name : str
            group name
        as_json : bool
            return result as json format

        Returns
        -------
        group: :class:Group
            group object

        """
        url = client.build_url('groups', group_id)
        body = {"name": name, }
        _, res_json = client.put(url, body=body, headers=self.headers)

        if as_json or not res_json:
            return as_json

        return Group(**res_json)

    def create(self, name, as_json=False):
        """Create new group.

        https://developers.mailerlite.com/v2/reference#create-group

        Parameters
        ----------
        name : str
            group name
        as_json : bool
            return result as json format

        Returns
        -------
        group: :class:Group
            group object

        """
        url = client.build_url('groups')
        data = {'name': name}
        _, res_json = client.post(url, body=data, headers=self.headers)

        if as_json or not res_json:
            return res_json

        return Group(**res_json)

    def add_subscribers(self, group_id, subscribers_data, resubscribe=False,
                        autoresponders=False, as_json=False):
        """Add one or many new subscribers to specified group at once.

        https://developers.mailerlite.com/v2/reference#add-single-subscriber
        https://developers.mailerlite.com/v2/reference#add-many-subscribers

        Parameters
        ----------
        group_id : int
            group id
        subscribers_data : dict, list of dict
            subscribers element that contains email and name
        resubscribe : bool
            reactivate subscriber if value is true (default False)
        autoresponders : bool
            autoresponders will be sent if value is true (default False)
        as_json : bool
            return result as json format

        Returns
        -------
        group: :class:Group
            group object

        """
        url = client.build_url('groups', group_id, 'subscribers', 'import')

        body = {'resubscribe': resubscribe, 'autoresponders': autoresponders}
        if isinstance(subscribers_data, dict):
            body['subscribers'] = [subscribers_data, ]
        elif isinstance(subscribers_data, list):
            body['subscribers'] = subscribers_data
        else:
            raise ValueError('subscribers_data should be a dict or a list of'
                             ' dict that contains the following keys: email,'
                             ' name')

        errors = [d for d in body['subscribers']
                  if not {'email', 'name'}.issubset(d.keys())]
        if errors:
            raise ValueError('All subscribers_data should contain the'
                             ' following keys: email, name')
        _, res_json = client.post(url, body=body, headers=self.headers)

        if as_json or not res_json:
            return res_json

        return [Subscriber(**subs) for subs in res_json['imported']]

    def add_single_subscriber(self, group_id, subscribers_data: dict,
                              resubscribe=False, autoresponders=False,
                              as_json=False):
        """Add single new subscriber to specified group.

        https://developers.mailerlite.com/v2/reference#add-single-subscriber

        Parameters
        ----------
        group_id : int
            group id
        subscribers_data : dict,
            subscribers data
        resubscribe : bool
            reactivate subscriber if value is true (default False)
        autoresponders : bool
            autoresponders will be sent if value is true (default False)
        as_json : bool
            return result as json format

        Returns
        -------
        subscriber: :class:Subscriber
            subscriber object

        """
        url = client.build_url('groups', group_id, 'subscribers')

        body = {'resubscribe': resubscribe, 'autoresponders': autoresponders,
                **subscribers_data}

        if not {'email', 'name'}.issubset(subscribers_data.keys()):
            raise ValueError('Subscribers_data should contain the'
                             ' following keys: email, name')

        _, res_json = client.post(url, body=body, headers=self.headers)

        if as_json or not res_json:
            return res_json

        return Subscriber(**res_json)

    def subscribers(self, group_id, limit=100, offset=0, stype=None,
                    as_json=False):
        """Get all subscribers in a specified group.

        https://developers.mailerlite.com/v2/reference#subscribers-in-a-group

        Parameters
        ----------
        group_id : int
            group id
        limit : int
            How many subscribers you want
        offset : int
            page index
        stype : str
            Define subscriber type: Here are the possible values:
            * None - All subscribers (default)
            * active
            * unsubscribed
            * bounced
            * junk
            * unconfirmed
        as_json : bool
            return result as json format

        Returns
        -------
        subscribers: list
            all desired Subscribers. More informations :
            https://developers.mailerlite.com/v2/reference#subscribers

        """
        params = {'limit': limit, 'offset': offset}
        if stype and stype.lower() in ['active', 'unsubscribed', 'bounced',
                                       'junk', 'unconfirmed']:
            params.update({'type': stype})

        url = client.build_url('groups', group_id, 'subscribers', **params)
        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        for res in res_json:
            res['fields'] = [Field(**field) for field in res['fields']]

        all_subscribers = [Subscriber(**res) for res in res_json]
        return all_subscribers

    def subscriber(self, group_id, subscriber_id, as_json=False):
        """Get one subscriber in a specified group.

        https://developers.mailerlite.com/v2/reference#a-subscriber-of-a-group

        Parameters
        ----------
        group_id : int
            group id
        subscriber_id : int
            subscriber id
        as_json : bool
            return result as json format

        Returns
        -------
        subscriber: :class:Subscriber
            a single subscriber
        """
        url = client.build_url('groups', group_id, 'subscribers',
                               subscriber_id)
        _, res_json = client.get(url, headers=self.headers)

        if as_json or not res_json:
            return res_json

        res_json['fields'] = [Field(**res) for res in res_json['fields']]

        return Subscriber(**res_json)

    def delete_subscriber(self, group_id, subscriber_id):
        """Remove a subscribers.

        Parameters
        ----------
        group_id : int
            group id
        subscriber_id : int
            subscriber id

        Returns
        -------
        success: bool
            deletion status
        """
        url = client.build_url('groups', group_id, 'subscribers',
                               subscriber_id)
        return client.delete(url, headers=self.headers)



"""Get Account information."""

import mailerlite.client as client


class Account:

    def __init__(self, headers):
        """Initialize Account object.

        Parameters
        ----------
        headers : dict
            request header containing your mailerlite api_key.
            More information : https://developers.mailerlite.com/docs/request

        """
        valid_headers, error_msg = client.check_headers(headers)
        if not valid_headers:
            raise ValueError(error_msg)

        self.headers = headers

    def info(self):
        """Get account info.

        https://developers.mailerlite.com/v2/reference#account

        Returns
        -------
        info: dict
            all account information.

        """
        url = client.build_url('me')
        _, res_json = client.get(url, headers=self.headers)

        return res_json

    def stats(self):
        """Get basic stats for of account.

        e.g. subscribers, open/click rates and so on.

        https://developers.mailerlite.com/v2/reference#stats

        Returns
        -------
        stats: dict
            account stats

        """
        url = client.build_url('stats')
        _, res_json = client.get(url, headers=self.headers)

        return res_json

    def double_optin(self):
        """Retrieve the status double opt-in.

        https://developers.mailerlite.com/v2/reference#get-double-optin-status

        Returns
        -------
        success: bool
            deletion status

        """
        url = client.build_url('settings', 'double_optin')
        _, res_json = client.get(url, headers=self.headers)
        return res_json

    def set_double_optin(self, enable):
        """Enable/disabled double opt-in for API and integrations.

        https://developers.mailerlite.com/v2/reference#change-double-optin-status

        Parameters
        ----------
        enable : bool
            status

        Returns
        -------
        status : dict
            action result

        """
        url = client.build_url('settings', 'double_optin')
        body = {'enable': enable}
        _, res_json = client.post(url, body=body, headers=self.headers)

        return res_json



"""Module to tests Segment."""
import pytest

from mailerlite.constants import API_KEY_TEST
from mailerlite.segment import Segments


@pytest.fixture
def header():
    headers = {'content-type': "application/json",
               'X-MailerLite-ApiDocs': "true",
               'x-mailerlite-apikey': API_KEY_TEST
               }
    return headers


def test_wrong_headers(header):
    # test valid first
    try:
        _ = Segments(header)
    except ValueError:
        return

    headers_2 = {'content-type': "application/json",
                 'x-mailerlite-apikey': 'FAKE_KEY'
                 }
    headers_3 = {'content-type': "application/json",
                 }
    headers_4 = {'x-mailerlite-apikey': 'FAKE_KEY'
                 }

    with pytest.raises(OSError):
        segm = Segments(headers_2)
        segm.count()

    with pytest.raises(ValueError):
        segm = Segments(headers_3)

    with pytest.raises(ValueError):
        segm = Segments(headers_4)


def test_segments_error(header):
    try:
        segm = Segments(header)
    except ValueError:
        return

    with pytest.raises(OSError):
        segm.all(order='upper')


def test_segments_crud(header):
    try:
        segm = Segments(header)
    except ValueError:
        return

    all_segm, meta = segm.all()

    assert len(all_segm) == meta.pagination.count
    assert len(all_segm) == segm.count()

    all_segm, meta = segm.all(as_json=True)

    assert len(all_segm)
    assert 'pagination' in meta.keys()



"""Module to tests Webhook."""

import pytest

from mailerlite.constants import API_KEY_TEST, Webhook
from mailerlite.webhook import Webhooks


@pytest.fixture
def header():
    headers = {'content-type': "application/json",
               'X-MailerLite-ApiDocs': "true",
               'x-mailerlite-apikey': API_KEY_TEST
               }
    return headers


def test_wrong_headers(header):
    try:
        _ = Webhooks(header)
    except ValueError:
        return

    headers_2 = {'content-type': "application/json",
                 'x-mailerlite-apikey': 'FAKE_KEY'
                 }
    headers_3 = {'content-type': "application/json",
                 }
    headers_4 = {'x-mailerlite-apikey': 'FAKE_KEY'
                 }

    with pytest.raises(OSError):
        wh = Webhooks(headers_2)
        wh.all()

    with pytest.raises(ValueError):
        wh = Webhooks(headers_3)

    with pytest.raises(ValueError):
        wh = Webhooks(headers_4)


def test_webhook_error(header):
    try:
        wh = Webhooks(header)
    except ValueError:
        return

    with pytest.raises(OSError):
        wh.update(123456, "new_url", "new_event")


def test_webhook_crud(header):
    try:
        wh = Webhooks(header)
    except ValueError:
        return

    all_wh = wh.all()

    assert len(all_wh) > 0
    assert len(all_wh) == wh.count()

    first_wh = all_wh[0]
    current_wh = wh.get(first_wh.id)

    # assert isinstance(current_wh, Webhook)
    for f in current_wh._fields:
        assert current_wh._asdict().get(f) == first_wh._asdict().get(f)

    # Need to Find a free webhooks tools for unittest
    # import ipdb; ipdb.set_trace()
    # code, custom_wh = wh.create(first_wh.url, first_wh.event)
    # assert code in [200, 201]

    # import ipdb; ipdb.set_trace()
    # assert custom_wh.event == first_wh.event
    # assert custom_wh.url == first_wh.url
    # assert custom_wh.id != first_wh.id
